using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms;

using pdftron;
using pdftron.PDF;
using pdftron.Common;

namespace PDFRedactViewCS
{
	/// <summary>
	/// MyPDFView is a sample class derived from pdftron.PDF.PDFView 
	/// that illustrates how to implement various custom tools and custom 
	/// drawing on top of PDFView. If you only need to view PDF documents 
	/// you can completely comment-out this class and instantiate 
	/// pdftron.PDF.PDFView directly.
	/// 
	/// MyPDFView shows how to implement various PDF tools (e.g. a tool 
	/// to create new link annotations, rectangular zoom, markup annotations, 
	/// link navigation, etc).
	/// </summary>
	public class MyPDFView : PDFView 
	{
		public enum CustomToolMode 
		{
			e_pan,                 // Pan tool
			e_text_rect_select,    // Rectangular text selection tool
			e_text_struct_select,   // HTML Style text selection tool
			e_draw,                // Markup tool
			e_link_create,         // Link-create tool
			e_zoomin,              // Rectangular zoom-in tool
			e_zoomout,              // Rectangular zoom-out tool
			// <-- Add new tool modes 
			e_rect_select
		}

		public static CustomToolMode _tool_mode = CustomToolMode.e_pan;


		public MyPDFView(PDFViewForm parent) : base() 
		{			
			_parent = parent;			
			_popup.Popup += new EventHandler(PopupEventHandler);
			this.ContextMenu = _popup;
			
			// this.ForeColor = Color.White;
			// this.BackColor = Color.White;
			// this.SetPageBorderVisibility(false);
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			// Draw the page as a background.
			base.OnPaint(e);

			if ((_tool_mode == CustomToolMode.e_link_create ||
				 _tool_mode == CustomToolMode.e_zoomin || 
				 _tool_mode == CustomToolMode.e_zoomout)
				&& Capture && _cur_page>0) 
			{
				// Draw Link rectangle annotations on top of the PDF page.
				float top, bottom, left, right;
				if (_start_pt.Y < _end_pt.Y) 
				{
					top = _start_pt.Y;
					bottom = _end_pt.Y;
				}
				else 
				{
					bottom = _start_pt.Y;
					top = _end_pt.Y;
				}

				if (_start_pt.X < _end_pt.X) 
				{
					left = _start_pt.X;
					right = _end_pt.X;
				}
				else 
				{
					right = _start_pt.X;
					left = _end_pt.X;
				}
				
				e.Graphics.DrawRectangle(Pens.Red, left, top, right-left, bottom-top);
			}
			else if ((_tool_mode == CustomToolMode.e_rect_select)&& _cur_page > 0)
			{
				if (_start_pt.X != _end_pt.X && _start_pt.Y != _end_pt.Y) // TODO add a fuzz compare
				{
					// Draw  highlight rectangle  on top of the PDF page.
					float top = Math.Min(_start_pt.Y, _end_pt.Y);
					float bottom = Math.Max(_start_pt.Y, _end_pt.Y);
					float left = Math.Min(_start_pt.X, _end_pt.X);
					float right = Math.Max(_start_pt.X, _end_pt.X);
					float height = Math.Abs(bottom - top);
					float width = Math.Abs(right  - left);
					e.Graphics.DrawRectangle(Pens.Blue, left, top, width, height);
				}
			}
			
			if (_freehand_markup != null) // Draw FreeHand annotations on top.
			{   
				PagePresentationMode pmode = GetPagePresentationMode();
				int curpage = GetCurrentPage();
				
				Graphics g = e.Graphics;
				foreach( DictionaryEntry itr in _freehand_markup)
				{
					int annot_page_num = (int) itr.Key;
					if ((pmode == PDFView.PagePresentationMode.e_single_page && annot_page_num!=curpage) || 
						(pmode == PDFView.PagePresentationMode.e_facing && annot_page_num!=curpage && annot_page_num!=curpage+1))
					{
						continue;
					}

					GraphicsState gs = g.Save();
					Matrix page2scr_mtx = GetDeviceTransform(annot_page_num);
					g.MultiplyTransform(page2scr_mtx);
					FreeHandAnnot annot = (FreeHandAnnot) itr.Value;
					annot.DrawAnnots(g, Pens.Red);
					g.Restore(gs);
				}
			}
		}

		protected override void OnPaintBackground(PaintEventArgs pevent) {}


		protected PointF _start_pt = new PointF(0,0);
		protected PointF _end_pt = new PointF(0,0);
		protected int _cur_page = 0;


		public void MyClearSelection()
		{
			ClearSelection();
			_start_pt = new PointF(0,0);
			_end_pt = new PointF(0,0);
		}


		// A dictionary mapping page numbers to FreeHandAnnot-s (used to implement freehand markup tool)
		public Hashtable _freehand_markup = null;
		protected GraphicsPath _curstroke = null;  // Used to track the current stroke.
	
		protected override void OnMouseDown(MouseEventArgs e) 
		{
			if (_tool_mode == CustomToolMode.e_link_create ||
				_tool_mode == CustomToolMode.e_zoomin || 
				_tool_mode == CustomToolMode.e_zoomout
				) 
			{
				_cur_page = GetPageNumberFromScreenPt(e.X, e.Y);
				if (_cur_page < 1) 
				{
					_cur_page = GetCurrentPage();
				}

				// Record the position for the rectangle origin.
				_start_pt.X = e.X; _start_pt.Y = e.Y;
				_end_pt.X = e.X;
				_end_pt.Y = e.Y;
				Capture = true;
			}
			else if (_tool_mode == CustomToolMode.e_draw ) 
			{
				// Sample code illustrating how to draw new content on top 
				// of PDF page...
				_cur_page = GetPageNumberFromScreenPt(e.X, e.Y);
				if (_cur_page < 1) 
				{
					_cur_page = GetCurrentPage();
				}

				
				_curstroke = new GraphicsPath();  // Start a new stroke.

				// Record the current position.
				double x=e.X, y=e.Y;
				this.ConvScreenPtToPagePt(ref x, ref y, _cur_page);
				_start_pt.X = (float)x; _start_pt.Y = (float)y;
				Capture = true;
			}
			if (_tool_mode == CustomToolMode.e_rect_select)
			{
				_cur_page = GetPageNumberFromScreenPt(e.X, e.Y);
				if (_cur_page < 1)
				{
					_cur_page = GetCurrentPage();
				}

				double x = e.X, y = e.Y;
//				this.ConvScreenPtToPagePt(ref x, ref y, _cur_page);
				// Record the position for the rectangle origin.
				_start_pt.X = (float)x; _start_pt.Y = (float)y;
				_end_pt.X = (float)x; _end_pt.Y = (float)y;
				Capture = true;
			}
			else 
			{
				// Sample code illustrating how to implement link navigation.
				if (IsFinishedRendering())
				{
					int page_num = GetPageNumberFromScreenPt(e.X, e.Y);
					if (page_num<1) return;

					// Find the click point in page coordinate system...
					double x = e.X, y = e.Y;
					ConvScreenPtToPagePt(ref x, ref y, page_num);

					Page page = GetDoc().GetPage(page_num);
					int annot_num = page.GetNumAnnots();
					for (int i=0; i<annot_num; ++i) 
					{
						Annot annot = page.GetAnnot(i);
						// Process only link annotations...
						if (annot.IsValid() == false || 
							annot.GetType() != Annot.Type.e_Link) continue;

						Rect box = annot.GetRect();
						if (box.Contains(x, y))
						{
							// Execute the action - jump to the destination page 
							Action action = annot.GetLinkAction(); 
							if (action.IsValid() == false) continue;
							if (action.GetType() == Action.Type.e_GoTo) 
							{
								Destination dest = action.GetDest();
								if (dest.IsValid())
								{
									page_num = dest.GetPage().GetIndex(); 
									SetCurrentPage(page_num);
									return;
								}
							}
							else if (action.GetType() == Action.Type.e_URI) 
							{
								// open the browser and jump to the given hyperlink...
								string target;
								
								target=action.GetSDFObj().Get("URI").Value().GetAsPDFText();
								try
								{
									System.Diagnostics.Process.Start(target);
								}
								catch(System.ComponentModel.Win32Exception noBrowser) 
								{
									if (noBrowser.ErrorCode==-2147467259)
										MessageBox.Show(noBrowser.Message);
								}
								catch (System.Exception other)
								{
									MessageBox.Show(other.Message);
								}
								break;
							}
							else if (action.GetType() == Action.Type.e_GoToR) 
							{
								// if the remote document is PDF, possibly open the
								// document as a new PDFViewForm...
							}
						}					
					}
				}
			}
		}

		protected override void OnMouseMove(MouseEventArgs e) 
		{
			if ((_tool_mode == CustomToolMode.e_link_create ||
				 _tool_mode == CustomToolMode.e_zoomin || 
				 _tool_mode == CustomToolMode.e_zoomout)		
				&& Capture && _cur_page>0) 
			{
				_end_pt.X = e.X; _end_pt.Y = e.Y;
				Invalidate();
			}
			else if (_tool_mode == CustomToolMode.e_pan)
			{   
				// Sample code showing how to change mouse cursor when hovering
				// over link annotations (in pan mode)...
				if (IsFinishedRendering())
				{
					int page_num = GetPageNumberFromScreenPt(e.X, e.Y);
					if (page_num<1) return;

					// Find the point in PDF page coordinate system...
					double x = e.X, y = e.Y;
					ConvScreenPtToPagePt(ref x, ref y, page_num);

					Page page = GetDoc().GetPage(page_num);
					int annot_num = page.GetNumAnnots();
					bool over_link = false;
					for (int i=0; i<annot_num; ++i)  
					{
						Annot annot = page.GetAnnot(i);
						if (annot.IsValid() == false || 
							annot.GetType() != Annot.Type.e_Link) continue;

						Rect box = annot.GetRect();
						if (box.Contains(x, y)) 
						{
							over_link = true;
							break;
						}
					}

					if (over_link) Cursor = Cursors.Cross;
					else Cursor = Cursors.Hand;
				}
			}
			else if (_tool_mode == CustomToolMode.e_draw && Capture && _curstroke!=null)
			{
				double x=e.X, y=e.Y;
				this.ConvScreenPtToPagePt(ref x, ref y, _cur_page);
				_end_pt.X = (float)x; _end_pt.Y = (float)y;
				_curstroke.AddLine(_start_pt.Y, _start_pt.X, _end_pt.Y, _end_pt.X);
				_start_pt.X = _end_pt.X; _start_pt.Y = _end_pt.Y;
				Invalidate();
			}
			else if (_tool_mode == CustomToolMode.e_rect_select)
			{
				if (Capture)
				{
					double x = e.X, y = e.Y;
//					this.ConvScreenPtToPagePt(ref x, ref y, _cur_page);
					_end_pt.X = (float)x; _end_pt.Y = (float)y;
					this.Cursor = System.Windows.Forms.Cursors.Cross;
					Invalidate();
				}
				else
				{
					this.Cursor = System.Windows.Forms.Cursors.Cross;
				}
			}
		}	

		protected override void OnMouseUp(MouseEventArgs e) 
		{
			if (_tool_mode == CustomToolMode.e_link_create && Capture && _cur_page>0) 
			{
				// The following sample code illustrates an implementation of link annotation 
				// tool that can be used to create new links on top of existing PDF documents.

				_end_pt.X = e.X; _end_pt.Y = e.Y;
			
				LinkCreate dlg = new LinkCreate();
				if (dlg.ShowDialog() == DialogResult.OK)
				{
					PDFDoc doc = GetDoc();
					Page pg = doc.GetPage(_cur_page);
					if (pg != null)
					{
						if (dlg.link.Text == "")
						{
							MessageBox.Show("The link text was not entered.", "PDFView", MessageBoxButtons.OK);
						}
						else
						{
							Action action = null;
							if (dlg.goto_webpage.Checked)
							{
								if (dlg.link.Text.StartsWith("http://"))
								{
									// Create a URI action
									pdftron.SDF.Obj a = doc.CreateIndirectDict();
									a.PutName("S", "URI");
									a.PutString("URI", dlg.link.Text);
									action = new Action(a);
								}
								else
								{
									MessageBox.Show("The hyperlink must start with 'http://'", "PDFView", MessageBoxButtons.OK);
								}
							}
							else
							{
								try
								{
									int page_num = Convert.ToInt32(dlg.link.Text);
									if (page_num < 1 || page_num > doc.GetPageCount())
									{
										MessageBox.Show("Invalid page number.", "PDFView", MessageBoxButtons.OK);
									}
									else
									{
										Page dest_page = doc.GetPage(page_num);
										action = Action.CreateGoto(Destination.CreateFitH(dest_page, dest_page.GetPageHeight()));
									}
								}
								catch (Exception)
								{
									MessageBox.Show("Invalid page number.", "PDFView", MessageBoxButtons.OK);
								}
							}

							if (action != null)
							{
								// Create link annotation position rectangle in PDF page coordinate system.

								Matrix screen2page = GetDeviceTransform(_cur_page);
								screen2page.Invert();

								PointF[] pt_arr = { _start_pt, _end_pt };
								screen2page.TransformPoints(pt_arr);

								Rect pos = new Rect(pt_arr[0].Y, pt_arr[0].X, pt_arr[1].Y, pt_arr[1].X);
								pos.Normalize();

								Annot link = Annot.CreateLink(doc, pos, action);

								// Todo? 
								// Create link appearances based on user preferences.
								// Create other annotation and action types...
								// ...

								pg.AnnotPushBack(link);
							}
						}
					}
				}

				Capture = false;
				_cur_page = 0;
				Invalidate();
			}
			else if ((_tool_mode == CustomToolMode.e_zoomin  || 
					  _tool_mode == CustomToolMode.e_zoomout) 
				&& Capture && _cur_page>0) 
			{
				// The following sample code illustrates an implementation of rectangular 
				// zoom-in and zoom-out tool (similar to the one in Acrobat Reader).

				_end_pt.X = e.X; _end_pt.Y = e.Y;
				double zoom_width = _end_pt.X - _start_pt.X; 
				double zoom_height = _end_pt.Y - _start_pt.Y; 

				if (Math.Abs(zoom_width)>20 && Math.Abs(zoom_height)>20) 
				{
					double z1, z2;
					if (_tool_mode == CustomToolMode.e_zoomin) 
					{
						z1 = GetBufferWidth()/Math.Abs(zoom_width); 
						z2 = GetBufferHeight()/Math.Abs(zoom_height); 
					}
					else  
					{
						z1 = Math.Abs(zoom_width)/GetBufferWidth();
						z2 = Math.Abs(zoom_height)/GetBufferHeight(); 
					}	
				
					double z = Math.Min(z1, z2) * GetZoom();
					SetZoom((int)(_start_pt.X + zoom_width/2.0), 
						(int)(_start_pt.Y + zoom_height/2.0), z);
				}

				Capture = false;
				_cur_page = 0;
				Invalidate();
			}
			else if (_tool_mode == CustomToolMode.e_draw && Capture)
			{   // Freehand tool
				double x=e.X, y=e.Y;
				this.ConvScreenPtToPagePt(ref x, ref y, _cur_page);
				_curstroke.AddLine(_start_pt.Y, _start_pt.X, (float)y, (float)x);

				Invalidate();
				_curstroke = null;
				Capture = false;
				Invalidate();
			}
			else if (_tool_mode == CustomToolMode.e_rect_select && Capture)
			{
				double x = e.X;
				double y = e.Y;
				_end_pt.X = (float)x;
				_end_pt.Y = (float)y;
				Capture = false;
				Invalidate();
			}
		}

		protected override void OnMouseWheel(MouseEventArgs e)
		{
			if (e.Delta > 0)
				GotoPreviousPage();
			else if (e.Delta < 0)
				GotoNextPage();
		}

		private bool IntersectsWith(Rect ra, Rect rb)
		{
			if (ra.x1 < rb.x2 &&
				ra.x2 >= rb.x1 &&
				ra.y1 < rb.y2 &&
				ra.y2 >= rb.y1)
			{
				return true;
			}
			return false;

		}

		public void Excise()
		{
			if (HasSelection())
			{
				ElementBuilder eb = new ElementBuilder();
				Selection sel = GetSelection();
				double[] selRects = sel.GetRects();
				if (selRects == null || selRects.Length == 0)
				{
					return;
				}
				double[] points = sel.GetRects();

				List<Rect> boundsRects = new List<Rect>();
				for (int idx = 0; idx <= points.Length - 1; idx += 4)
				{
					boundsRects.Add(new Rect(points[idx], points[idx + 1], points[idx + 2], points[idx + 3]));
				}



				PDFDoc doc = GetDoc();
				int curpage = GetCurrentPage();
				Page page = doc.GetPage(curpage);


				using (ElementWriter pageWriter = new ElementWriter())
				{
					using (ElementReader pageReader = new ElementReader())
					{
						Element element;

						pageReader.Begin(page);
						Page new_page = doc.PageCreate();
						pageWriter.Begin(new_page);

						while ((element = pageReader.Next()) != null) 	// Read page contents
						{
							if (element.GetType() == Element.Type.e_text)
							{
								Rect bbox = new Rect();
								element.GetBBox(bbox);
								RectangleOverlap overlap = GetOverlap(boundsRects, bbox);

								if (overlap == RectangleOverlap.None)
								{
									pageWriter.WriteElement(element);
								}
								else if (overlap == RectangleOverlap.Complete)
								{
									;// drop element
								}
								else
								{
									int charRunIndex = 0;
									Matrix2D ctm = element.GetCTM();
									Matrix2D Tm = element.GetTextMatrix();
									Matrix2D mtx = ctm * Tm;
									GState gs = element.GetGState();
									pdftron.PDF.Font font = gs.GetFont();
									Matrix2D mgx = font.GetType3FontMatrix();
									double fontSize = gs.GetFontSize();
									double unitsPerEm = font.GetUnitsPerEm();
									Rect bbox1 = font.GetBBox();
									double bbox_height = bbox1.y2 - bbox1.y1;
									double decent = fontSize * bbox1.y1 / bbox_height;
									double ascent = fontSize * bbox1.y2 / bbox_height;
									double horiz_scale = gs.GetHorizontalScale() / 100.0;
									bool first = true;
									Matrix2D startTm = new Matrix2D(Tm);
									double positionAdjustment = element.GetPosAdjustment();
									double offsetX = 0;
									double offsetY = 0;
									element.GetNewTextLineOffset(ref offsetX, ref offsetY);
									double textLength = element.GetTextLength();
									double textHeight = 0;
									mtx.Mult(ref textLength, ref textHeight);

									pageWriter.WriteElement(eb.CreateGroupBegin());

									for (CharIterator itr = element.GetCharIterator(); itr.HasNext(); itr.Next(), charRunIndex++)
									{
										CharData charData = itr.Current();
										Rect gbox = GetGlyphBBox(itr, font, horiz_scale, fontSize, ascent, decent);
										gbox = GetBBoxTransfRect(gbox, mtx);
										if (GetOverlap(boundsRects, gbox) == RectangleOverlap.None)
										{
											Element new_element = eb.CreateTextRun(new String((char)(itr.Current().char_code), 1), font, fontSize);
											Matrix2D newTm = new Matrix2D(startTm);
											double h = charData.x;
											double v = charData.y;
											mtx.Mult(ref h, ref v);	

											newTm.m_h = h;
											newTm.m_v = v;
											new_element.SetTextMatrix(newTm);

											if (first)
											{
												new_element.SetPosAdjustment(element.GetPosAdjustment());
												first = false;
											}
											GState ngs = new_element.GetGState();
											ngs.SetBlackGenFunct(gs.GetBlackGenFunct());
											ngs.SetBlendMode(gs.GetBlendMode());
											ngs.SetCharSpacing(gs.GetCharSpacing());
											ngs.SetDashPattern(gs.GetDashes(), gs.GetPhase());
											ngs.SetFillColorSpace(gs.GetFillColorSpace());
											ngs.SetFillOpacity(gs.GetFillOpacity());
											ngs.SetFlatness(gs.GetFlatness());
											ngs.SetHalftone(gs.GetHalftone());
											ngs.SetHorizontalScale(gs.GetHorizontalScale());
											ngs.SetLeading(gs.GetLeading());
											ngs.SetLineCap(gs.GetLineCap());
											ngs.SetLineJoin(gs.GetLineJoin());
											ngs.SetLineWidth(gs.GetLineWidth());
											ngs.SetMiterLimit(gs.GetMiterLimit());
											ngs.SetRenderingIntent(gs.GetRenderingIntent());
											ngs.SetSmoothnessTolerance(gs.GetSmoothnessTolerance());
											ngs.SetSoftMask(gs.GetSoftMask());
											ngs.SetStrokeColor(gs.GetStrokeColor());
											ngs.SetStrokeColorSpace(gs.GetStrokeColorSpace());
											ngs.SetStrokeOpacity(gs.GetStrokeOpacity());
											ngs.SetTextKnockout(gs.IsTextKnockout());
											ngs.SetTextRenderMode(gs.GetTextRenderMode());
											ngs.SetTextRise(gs.GetTextRise());
											ngs.SetUCRFunct(gs.GetUCRFunct());
											ngs.SetWordSpacing(gs.GetWordSpacing());
											new_element.UpdateTextMetrics();
											pageWriter.WriteElement(new_element);
										}
									}
									pageWriter.WriteElement(eb.CreateGroupEnd());
									byte[] textData = new byte[1];
									textData[0] = 0;
									element.SetTextData(textData, 1);
									element.SetTextMatrix(startTm);
									element.UpdateTextMetrics();
									pageWriter.WriteElement(element); 
								}
							}
							else
							{
								pageWriter.WriteElement(element);
							}
						}
						pageWriter.End();
						pageReader.End();
						new_page.SetMediaBox(page.GetCropBox());
						doc.PagePushBack(new_page);
						doc.PageRemove(doc.GetPageIterator(curpage)); // Remove the old page
					}
				}
			}

			MyClearSelection();
			Update();
			Invalidate();
		}



		public static class RectHelper
		{
			public static bool IntersectsWith(Rect r1, Rect r2)
			{
				if ((r1.x1 < r2.x2 && r1.x2 > r2.x1)
					&& (r1.y1 < r2.y2 && r1.y2 > r2.y1))
				{
					return true;
				}
				return false;
			}
		}

	  static Rect GetBBoxTransfRect(Rect inn, Matrix2D mtx)
	  {
	   double p1x = inn.x1, p1y = inn.y1, p2x = inn.x2, p2y = inn.y1, p3x = inn.x2, p3y = inn.y2, p4x = inn.x1, p4y = inn.y2;
	   mtx.Mult(ref p1x, ref p1y);
	   mtx.Mult(ref p2x, ref p2y);
	   mtx.Mult(ref p3x, ref p3y);
	   mtx.Mult(ref p4x, ref p4y);

	   return new Rect(Math.Min(Math.Min(Math.Min(p1x, p2x), p3x), p4x),
		  Math.Min(Math.Min(Math.Min(p1y, p2y), p3y), p4y),
		  Math.Max(Math.Max(Math.Max(p1x, p2x), p3x), p4x),
		  Math.Max(Math.Max(Math.Max(p1y, p2y), p3y), p4y));
	  }

	  private static RectangleOverlap GetOverlap(List<Rect> boundingRectangles, Rect innerRectangle)
	  {
	   RectangleOverlap greatestOverlap = RectangleOverlap.None;

	   foreach (Rect boundingRectangle in boundingRectangles)
		greatestOverlap = (RectangleOverlap)Math.Max((int)greatestOverlap, (int)GetOverlap(boundingRectangle, innerRectangle));

	   return greatestOverlap;
	  }

	  private static RectangleOverlap GetOverlap(Rect outerRectangle, Rect innerRectangle)
	  {

	   if (innerRectangle.x2 <= outerRectangle.x1 || innerRectangle.x1 >= outerRectangle.x2 || innerRectangle.y1 >= outerRectangle.y2 || innerRectangle.y2 <= outerRectangle.y1)
	   {
		return RectangleOverlap.None;
	   }
	   else if (innerRectangle.x2 <= outerRectangle.x2 && innerRectangle.x1 >= outerRectangle.x1 && innerRectangle.y1 >= outerRectangle.y1 && innerRectangle.y2 <= outerRectangle.y2)
	   {
		return RectangleOverlap.Complete;
	   }
	   else
	   {
		return RectangleOverlap.Partial;
	   }
	  }

	  private enum RectangleOverlap
	  {
	   None = 1,
	   Partial = 2,
	   Complete = 3
	  }

		static Rect GetGlyphBBox(CharIterator itr, pdftron.PDF.Font font, double horiz_scale, double font_sz, double ascent, double descent)
		{
			Rect out_bbox = new Rect();
			out_bbox.x1 = itr.Current().x;
			out_bbox.y1 = itr.Current().y;

			double dx = 0;
			if (font.IsSimple())
			{
				dx = font.GetWidth(itr.Current().char_code) / 1000.0;
				dx *= horiz_scale * font_sz;

				out_bbox.x2 = out_bbox.x1 + dx;
				out_bbox.y1 = itr.Current().y + descent;
				out_bbox.y2 = itr.Current().y + ascent;
			}
			else
			{
				int cid = font.MapToCID(itr.Current().char_code);
				dx = font.GetWidth(cid) / 1000.0;
				dx *= horiz_scale * font_sz;

				out_bbox.x2 = out_bbox.x1 + dx;
				out_bbox.y1 = itr.Current().y + descent;
				out_bbox.y2 = itr.Current().y + ascent;
			}

			return out_bbox;
		}



	
		public void Mark()
		{
			try
			{
				if (_tool_mode == CustomToolMode.e_text_rect_select ||
					_tool_mode == CustomToolMode.e_text_struct_select )
				{
					if (HasSelection())
					{
						Selection sel = GetSelection();
						double[] selRects = sel.GetRects();
						PDFDoc doc = GetDoc();
						int curpage = GetCurrentPage();
						Page pg = doc.GetPage(curpage);

						using (ElementBuilder bld = new ElementBuilder())	// Used to build new Element objects
						{
							using (ElementWriter writer = new ElementWriter())	// Used to write Elements to the page
							{
								writer.Begin(pg, false);				// Begin writing to this page
								for (int i = 0; i < selRects.GetLength(0); )
								{
									double left = selRects[i++];
									double top = selRects[i++];
									double right = selRects[i++];
									double bottom = selRects[i++];
									double width = Math.Abs(right - left);
									double height = Math.Abs(top - bottom);


									System.Drawing.Bitmap bmp = new System.Drawing.Bitmap((int)Math.Ceiling(width), (int)Math.Ceiling(height),
										System.Drawing.Imaging.PixelFormat.Format32bppArgb);
									for (int x0 = 0; x0 < width; x0++)
									{
										for (int y0 = 0; y0 < height; y0++)
										{
											bmp.SetPixel(x0, y0, Color.LightSkyBlue);
										}
									}
									pdftron.PDF.Image img = pdftron.PDF.Image.Create(doc, bmp);
									Element element = bld.CreateImage(img, left, top, width, height);
									writer.WritePlacedElement(element);
									writer.Flush();
								}
								writer.End();  // save changes to the current page
							}
						}
					}
					MyClearSelection();
					Update();
					Invalidate();
				}
				else if (_tool_mode == CustomToolMode.e_rect_select)
				{
					double top = (double)Math.Min(_start_pt.Y, _end_pt.Y);
					double bottom = (double)Math.Max(_start_pt.Y, _end_pt.Y);
					double left = (double)Math.Min(_start_pt.X, _end_pt.X);
					double right = (double)Math.Max(_start_pt.X, _end_pt.X);
					this.ConvScreenPtToPagePt(ref left, ref top, _cur_page);
					this.ConvScreenPtToPagePt(ref right, ref bottom, _cur_page);
					// BUG this ignores the usual rotated page problems
					double height = Math.Abs(bottom - top);
					double width = Math.Abs(right - left);

					PDFDoc doc = GetDoc();
					Page pg = doc.GetPage(_cur_page);

					using (ElementBuilder bld = new ElementBuilder())	// Used to build new Element objects
					{
						using (ElementWriter writer = new ElementWriter())	// Used to write Elements to the page
						{
							writer.Begin(pg, false);				// Begin writing to this page


							// To do find a graphic
							System.Drawing.Bitmap bmp = new System.Drawing.Bitmap((int)Math.Ceiling(width), (int)Math.Ceiling(height),
								System.Drawing.Imaging.PixelFormat.Format32bppArgb);
							for (int x = 0; x < width; x++)
							{
								for (int y = 0; y < height; y++)
								{
									bmp.SetPixel(x, y, Color.LightYellow);
								}
							}

							pdftron.PDF.Image img = pdftron.PDF.Image.Create(doc, bmp);
							//Element source = bld.CreateImage(img, left, top, 1.0, 1.0);
							//							Element source = bld.CreateImage(img, left, top - height, 255, 150);
							Element element = bld.CreateImage(img, left, top - height, width, height);

							writer.WritePlacedElement(element);
							writer.Flush();
							writer.End();  // save changes to the current page
						}
					}
					MyClearSelection();
					Update();
					Invalidate();
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString());
			}

		}



		private ContextMenu _popup = new ContextMenu();

		private void PopupEventHandler(System.Object sender, System.EventArgs e)
		{			
			_popup.MenuItems.Clear();	// Clear all previously added MenuItems.

			if(HasSelection()) // Add MenuItems to display if there is a text selection.
			{				
				MenuItem menuItem1 = new MenuItem("Cop&y");
				menuItem1.Click += new EventHandler(OnTextCopy);
				_popup.MenuItems.Add(menuItem1);
			}

			if (_tool_mode == CustomToolMode.e_pan ||
				_tool_mode == CustomToolMode.e_text_rect_select ||
				_tool_mode == CustomToolMode.e_text_struct_select) 
			{
				MenuItem menuItem2 = new MenuItem("Se&lect All");
				menuItem2.Click += new EventHandler(OnSelectAll);
				_popup.MenuItems.Add(menuItem2);

				MenuItem menuItem3 = new MenuItem("D&eselect All");
				menuItem3.Click += new EventHandler(OnDeselectAll);
				_popup.MenuItems.Add(menuItem3);

			}
					
			{   // Context menu for all tool modes.
				//MenuItem menuItem1 = new MenuItem("Hand Tool");
				//menuItem1.Click += new EventHandler(OnSelectHandTool);
				//_popup.MenuItems.Add(menuItem1);

				//MenuItem menuItem2 = new MenuItem("Text Select Tool");
				//menuItem2.Click += new EventHandler(OnSelectTextTool);
				//_popup.MenuItems.Add(menuItem2);

				MenuItem menuItem3 = new MenuItem("Print");
				menuItem3.Click += new EventHandler(OnPrint);
				_popup.MenuItems.Add(menuItem3);
			}
		}

		public void OnTextCopy(object sender, EventArgs e)
		{
			Selection sel = GetSelection();
			// NET2 method: Clipboard.SetText(sel.GetText(), TextDataFormat.Html);
			// doesn't work in some applications, so we provide our own implementation:
			MyPDFView.CopyHtmlToClipBoard(sel.GetAsHtml(), sel.GetAsUnicode());
		}

		private void OnSelectAll(object sender, EventArgs e) {
			SelectAll();
		}

		private void OnDeselectAll(object sender, EventArgs e) {
			MyClearSelection();
		}

		// private void OnSelectTextTool(object sender, EventArgs e) { _parent.SetToolMode(CustomToolMode.e_text_rect_select); }
		// private void OnSelectHandTool(object sender, EventArgs e) {	_parent.SetToolMode(CustomToolMode.e_pan); }

		private void OnPrint(object sender, EventArgs e) 
		{
			_parent.Print();
		}

		private PDFViewForm _parent;

		// A utility method to copy HTML selection to clipboard.
		static public void CopyHtmlToClipBoard(string html, string txt) 
		{
			System.Text.Encoding enc = System.Text.Encoding.UTF8;

			string begin = "Version:0.9\r\nStartHTML:{0:000000}\r\nEndHTML:{1:000000}"
				+ "\r\nStartFragment:{2:000000}\r\nEndFragment:{3:000000}\r\n";

			string html_begin = "<html>\r\n<head>\r\n"
				+ "<meta http-equiv=\"Content-Type\""
				+ " content=\"text/html; charset=" + enc.WebName + "\">\r\n"
				+ "<title>HTML clipboard</title>\r\n</head>\r\n<body>\r\n"
				+ "<!--StartFragment-->";

			string html_end = "<!--EndFragment-->\r\n</body>\r\n</html>\r\n";

			string begin_sample = String.Format(begin, 0, 0, 0, 0);

			int count_begin = enc.GetByteCount(begin_sample);
			int count_html_begin = enc.GetByteCount(html_begin);
			int count_html = enc.GetByteCount(html);
			int count_html_end = enc.GetByteCount(html_end);

			string html_total = String.Format(
				begin
				, count_begin
				, count_begin + count_html_begin + count_html + count_html_end
				, count_begin + count_html_begin
				, count_begin + count_html_begin + count_html
				) + html_begin + html + html_end;

			DataObject obj = new DataObject();
			obj.SetData(DataFormats.Html, new System.IO.MemoryStream(
				enc.GetBytes(html_total)));

			obj.SetData(DataFormats.UnicodeText, txt);

			Clipboard.SetDataObject(obj, true);
		}

	}
}
